from re import search as re_search
from time import perf_counter
from time import sleep
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from .connection import CloudflareScraper
from .connection import Response
from .connection import get
from .connection import get_binary_raw
from .connection import get_robots
from .connection import join_url
from .connection import make_session
from .parse import BeautifulSoup
from .parse import page_parse
from .sub import Sub
from .sub import SubPartial


class FAAPI:
    def __init__(self, cookies: List[Dict[str, Any]] = None):
        self.cookies: List[Dict[str, Any]] = [] if cookies is None else cookies
        self.session: CloudflareScraper = make_session(self.cookies)
        self.robots: Dict[str, List[str]] = get_robots()
        self.crawl_delay: float = float(self.robots["Crawl-delay"][0]) if self.robots.get("Crawl-delay", 0) else 1.0
        self.last_get: float = perf_counter() - self.crawl_delay
        self.raise_for_delay: bool = False

    def handle_delay(self):
        if (delay_diff := perf_counter() - self.last_get) < self.crawl_delay:
            if self.raise_for_delay:
                raise Exception(f"Crawl-delay not respected {delay_diff}<{self.crawl_delay}")
            sleep(self.crawl_delay - delay_diff)
        self.last_get = perf_counter()

    def check_path(self, path: str):
        for pattern in self.robots.get("disallow", []):
            if re_search(pattern.replace("*", ".*").strip("/"), path):
                raise Exception(f"Path {path} is not allowed by robots.txt {pattern}")

    def get(self, path: str, **params) -> Response:
        self.check_path(path)
        self.handle_delay()
        return get(self.session, path, **params)

    def get_parse(self, path: str, **params) -> Optional[BeautifulSoup]:
        res = self.get(path, **params)
        return page_parse(res.text) if res.ok else None

    def get_sub(self, sub_id: int, get_file: bool = False) -> Tuple[Sub, Optional[bytes]]:
        assert isinstance(sub_id, int) and sub_id > 0

        sub_page = self.get_parse(f"/view/{sub_id}")
        sub = Sub(sub_page)
        sub_file = self.get_sub_file(sub) if get_file else None

        return sub, sub_file

    def get_sub_file(self, sub: Sub) -> Optional[bytes]:
        self.handle_delay()
        return get_binary_raw(self.session, sub.file_url)

    def userpage(self, user: str) -> Tuple[str, str, Optional[BeautifulSoup]]:
        assert isinstance(user, str)

        page: Optional[BeautifulSoup] = self.get_parse(join_url("user", user))
        if page is None:
            return "", "", None

        username_div = page.find(name="div", attrs={"class": "username"})

        username = username_div.find("span").text.strip()
        status = username[0]
        username = username[1:]

        description = page.find(name="div", attrs={"class": "userpage-profile"}).text.strip()

        return username, status, page_parse(description)

    def gallery(self, user: str, page: int = 1) -> Tuple[List[SubPartial], int]:
        assert isinstance(user, str)
        assert isinstance(page, int) and page >= 1

        page_parsed = self.get_parse(join_url("gallery", user, str(page)))

        if page_parsed is None or page_parsed.title.text.lower().startswith("account disabled"):
            return [], 0

        subs = page_parsed.findAll(name="figure")

        return list(map(SubPartial, subs)), (page + 1) if subs else 0

    def scraps(self, user: str, page: int = 1) -> Tuple[List[SubPartial], int]:
        assert isinstance(user, str)
        assert isinstance(page, int) and page >= 1

        page_parsed = self.get_parse(join_url("scraps", user, str(page)))

        if page_parsed is None or page_parsed.title.text.lower().startswith("account disabled"):
            return [], 0

        subs = page_parsed.findAll(name="figure")

        return list(map(SubPartial, subs)), (page + 1) if subs else 0

    def favorites(self, user: str, page: str = "") -> Tuple[List[SubPartial], str]:
        assert isinstance(user, str)
        assert isinstance(page, str)

        page_parsed = self.get_parse(join_url("favorites", user, page))

        if page_parsed is None or page_parsed.title.text.lower().startswith("account disabled"):
            return [], ""

        subs = page_parsed.findAll(name="figure")

        button_next = page_parsed.find(name="a", attrs={"class": "button standard right"})
        page_next: str = button_next["href"].split("/", 3)[-1] if button_next else ""

        return list(map(SubPartial, subs)), page_next

    def search(self, q: str, page: int = 1, **params) -> Tuple[List[SubPartial], int, int, int, int]:
        assert isinstance(q, str)
        assert isinstance(page, int)

        page_parsed = self.get_parse("search", q=q, page=page, **params)

        subs = page_parsed.findAll(name="figure")
        if page_parsed is None:
            return [], 0, 0, 0, 0

        query_stats = page_parsed.find("div", id="query-stats")
        for div in query_stats("div"):
            div.decompose()

        a, b, tot = map(int, re_search(r"(\d+)[^\d]*(\d+)[^\d]*(\d+)", query_stats.text.strip()).groups())
        page_next = (page + 1) if b < tot else 0

        return list(map(SubPartial, subs)), page_next, a, b, tot

    def user_exists(self, user: str):
        assert isinstance(user, str)

        res = self.get(join_url("view", user))

        if not res.ok:
            return False
        elif not (title := page_parse(res.text).title.text):
            return False
        elif title.text.lower() == "system error":
            return False
        elif title.text.lower().startswith("account disabled"):
            return False
        else:
            return True

    def sub_exists(self, sub_id: Union[int, str]):
        assert isinstance(sub_id, int) or (isinstance(sub_id, str) and sub_id.isdigit())

        res = self.get(join_url("view", str(sub_id)))

        if not res.ok:
            return False
        elif not (title := page_parse(res.text).title.text):
            return False
        elif title.text.lower() == "system error":
            return False
        else:
            return True



from time import sleep
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from cfscrape import CloudflareScraper
from cfscrape import create_scraper
from requests import Response
from requests import get as get_raw
from requests.exceptions import ConnectTimeout
from requests.exceptions import ConnectionError

root = "https://www.furaffinity.net"


def join_url(*url_comps: str) -> str:
    return '/'.join(map(lambda e: e.strip(" /"), url_comps))


def ping() -> bool:
    try:
        return get_raw(root).ok
    except (ConnectionError, ConnectTimeout):
        return False


def make_session(cookies: List[dict]) -> Optional[CloudflareScraper]:
    if not ping():
        return None

    cookies = [
        {"name": cookie["name"], "value": cookie["value"]}
        for cookie in cookies
        if "name" in cookie and "value" in cookie
    ]

    session: CloudflareScraper = create_scraper()

    for cookie in cookies:
        session.cookies.set(cookie["name"], cookie["value"])

    return session


def get_robots() -> Dict[str, List[str]]:
    res = get_raw(join_url(root, "robots.txt"))

    if not res.ok:
        return {}

    robot: Dict[str, Union[int, str, List[str]]] = {}

    for elem in filter(lambda r: r not in ("", "#"), res.text.split("\n")):
        key, value = elem.split(":", 1)
        robot[key] = robot.get(key, []) + [value.strip()]

    return robot


def get(session: CloudflareScraper, path: str, **params) -> Response:
    return session.get(join_url(root, path), params=params)


def get_binary_raw(session: CloudflareScraper, url: str, speed: Union[int, float] = 100) -> Optional[bytes]:
    assert isinstance(speed, int) or isinstance(speed, float)

    file_stream = session.get(url, stream=True)

    if not file_stream.ok:
        file_stream.close()
        return None

    file_binary = bytes()
    for chunk in file_stream.iter_content(chunk_size=1024):
        file_binary += chunk
        sleep(1 / speed) if speed > 0 else None

    return file_binary



from re import search as re_search
from typing import List

from bs4 import BeautifulSoup
from bs4 import ResultSet
from bs4.element import Tag
from dateutil.parser import parse


class SubPartial:
    def __init__(self, figure_tag: Tag):
        self.id: int = 0
        self.title: str = ""
        self.author: str = ""
        self.rating: str = ""
        self.type: str = ""

        self.parse_figure_tag(figure_tag)

    def __iter__(self):
        yield "id", self.id
        yield "title", self.title
        yield "author", self.author
        yield "rating", self.rating
        yield "type", self.type

    def __repr__(self):
        return repr(dict(self))

    def parse_figure_tag(self, figure_tag: Tag):
        caption_links: ResultSet = figure_tag.find("figcaption").findAll("a")

        self.id = int(caption_links[0]["href"].strip("/").split("/")[-1])
        self.title = caption_links[0].text
        self.author = caption_links[1].text
        self.rating, self.type = figure_tag["class"][0][2:], figure_tag["class"][1][2:]


class Sub:
    def __init__(self, sub_page: BeautifulSoup = None):
        assert sub_page is None or isinstance(sub_page, BeautifulSoup)

        self.sub_page = sub_page

        self.id: int = 0
        self.title: str = ""
        self.author: str = ""
        self.date: str = ""
        self.tags: List[str] = []
        self.category: str = ""
        self.species: str = ""
        self.gender: str = ""
        self.rating: str = ""
        self.description: str = ""
        self.file_url: str = ""

        self.parse_page()

    def __iter__(self):
        yield "id", self.id
        yield "title", self.title
        yield "author", self.author
        yield "date", self.date
        yield "tags", self.tags
        yield "category", self.category
        yield "species", self.species
        yield "gender", self.gender
        yield "rating", self.rating
        yield "description", self.description
        yield "file_url", self.file_url

    def __repr__(self):
        return repr(dict(self))

    def parse_page(self):
        if self.sub_page is None:
            return

        if self.sub_page.find("section", attrs={"class": "notice-message"}):
            raise Exception("Error: notice-message section found")

        tag_id: Tag = self.sub_page.find("meta", property="og:url")
        tag_sub_info: Tag = self.sub_page.find("div", attrs={"class": "submission-id-sub-container"})
        tag_title: Tag = tag_sub_info.find("div", attrs={"class": "submission-title"})
        tag_author: Tag = tag_sub_info.find("a")
        tag_date: Tag = self.sub_page.find("span", attrs={"class": "popup_date"})
        tag_tags: ResultSet = self.sub_page.find("section", attrs={"class": "tags-row"}).findAll("a")
        tag_rating: Tag = self.sub_page.find("div", attrs={"class": "rating"}).find("span")
        tag_info: ResultSet = self.sub_page.find("section", attrs={"class": "info text"}).findAll("div")
        tag_category1: Tag = tag_info[1].find("span", attrs={"class": "category-name"})
        tag_category2: Tag = tag_info[1].find("span", attrs={"class": "type-name"})
        tag_species: Tag = tag_info[2].find("span")
        tag_gender: Tag = tag_info[3].find("span")
        tag_description: Tag = self.sub_page.find("div", attrs={"class": "submission-description"})
        tag_file_url: Tag = self.sub_page.find("div", attrs={"class": "download"}).find("a")

        self.id = int(tag_id["content"].strip("/").split("/")[-1])
        self.title = tag_title.text.strip()
        self.author = tag_author.text.strip()
        self.date = parse(tag_date["title"].strip()).strftime("%Y-%m-%d")
        self.tags = [t.text.strip() for t in tag_tags]
        self.category = tag_category1.text.strip() + "/" + tag_category2.text.strip()
        self.species = tag_species.text.strip()
        self.gender = tag_gender.text.strip()
        self.rating = tag_rating.text.strip()
        self.description = "".join(map(str, tag_description.children)).strip()
        self.file_url = "https:" + tag_file_url["href"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['faapi']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.1,<5.0.0',
 'cfscrape>=2.1.1,<3.0.0',
 'lxml>=4.5.2,<5.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'faapi',
    'version': '2.3.0',
    'description': 'Python module to implement API-like functionality for the FurAffinity.net website.',
    'long_description': '# FurAffinity API\n\n[![pypi version](https://img.shields.io/pypi/v/faapi.svg)](https://pypi.org/project/faapi/)\n[![supported Python version](https://img.shields.io/pypi/pyversions/faapi.svg)](https://pypi.org/project/faapi/)\n[![license](https://img.shields.io/pypi/l/faapi.svg)](https://pypi.org/project/faapi/)\n\nPython module to implement API-like functionality for the FurAffinity.net website\n\n## Requirements\n\nPython 3.8+ is necessary to run this module.\n\nThis module requires the following pypi modules:\n\n* [beautifulsoup4](https://www.crummy.com/software/BeautifulSoup/)\n* [cfscrape](https://github.com/Anorov/cloudflare-scrape)\n* [lxml](https://github.com/lxml/lxml/)\n* [requests](https://github.com/requests/requests)\n* [python-dateutil](https://github.com/dateutil/dateutil/)\n\n## Usage\n\nThe API is comprised of a main class `FAAPI` and two submission classes: `Sub` and `SubPartial`.\nOnce `FAAPI` is initialized its method can be used to crawl FA and return machine-readable objects.\n\n```python\nimport faapi\nimport json\n\napi = faapi.FAAPI()\nsub, sub_file = api.get_sub(12345678, get_file=True)\n\nprint(sub.id, sub.title, sub.author, f"{len(sub_file)/1024:02f}KiB")\n\nwith open(f"{sub.id}.json", "w") as f:\n    f.write(json.dumps(sub))\n\nwith open(sub.file_url.split("/")[-1], "wb") as f:\n    f.write(sub_file)\n\ngallery, _ = api.gallery("user_name", 1)\nwith open("user_name-gallery.json", "w") as f:\n    f.write(json.dumps(gallery))\n```\n\n### robots.txt\n\nAt init, the `FAAPI` object downloads the [robots.txt](https://www.furaffinity.net/robots.txt) file from FA to determine the `Crawl-delay` value set therein.\nIf not set, a value of 1 second is used.\n\nTo respect this value, the default behaviour of of the `FAAPI` object is to wait when a get request is made if the last request was performed more recently then the crawl delay value.\n\nSee under [FAAPI](#faapi) for more details on this behaviour.\n\nFurthermore, any get operation that points to a disallowed path from robots.txt will raise an exception. This check should not be circumvented and the developer of this module does not take responsibility for violations of the TOS of FurAffinity.\n\n### Cookies\n\nTo access protected pages, cookies from an active session are needed. These cookies must be given to the FAAPI object as a list of dictionaries, each containing a `name` and a `value` field. The cookies list should look like the following random example:\n\n```python\ncookies = [\n    {"name": "a", "value": "38565475-3421-3f21-7f63-3d341339737"},\n    {"name": "b", "value": "356f5962-5a60-0922-1c11-65003b703038"},\n]\n```\n\nOnly cookies `a` and `b` are needed.\n\nTo access session cookies, consult the manual of the browser used to login.\n\n*Note:* it is important to not logout of the session the cookies belong, otherwise they will no longer work.\n\n## FAAPI\n\nThis is the main object that handles all the calls to scrape pages and get submissions.\n\nIt holds 6 different fields:\n\n* `cookies: List[dict] = []` cookies passed at init\n* `session: CloudflareScraper` cfscrape session used for get requests\n* `robots: Dict[str, List[str]]` robots.txt values\n* `crawl_delay: float` crawl delay from robots.txt, else 1\n* `last_get: float` time of last get (not UNIX time, uses `time.perf_counter` for more precision)\n* `raise_for_delay: bool = False` if set to `True`, raises an exception if a get call is made before enough time has passed\n\n### Init\n\n`__init__(cookies: List[dict] = None)`\n\nThe class init has a single optional argument `cookies` necessary to read logged-in-only pages.\nThe cookies can be omitted and the API will still be able to access public pages.\n\n*Note:* Cookies must be in the format mentioned above in [#Cookies](#cookies).\n\n### Methods\n\n* `get(path: str, **params) -> requests.Response`<br>\nThis returns a response object containing the result of the get operation on the given url with the optional `**params` added to it (url provided is considered as path from \'https://www.furaffinity.net/\').\n\n* `get_parse(path: str, **params) -> bs4.BeautifulSoup`<br>\nSimilar to `get()` but returns the parsed  HTML from the normal get operation.\n\n* `get_sub(sub_id: int, get_file: bool = False) -> Tuple[Sub, Optional[bytes]]`<br>\nGiven a submission ID in either int or str format, it returns a `Sub` object containing the various metadata of the submission itself and a `bytes` object with the submission file if `get_file` is passed as `True`.\n\n* `get_sub_file(sub: Sub) -> Optional[bytes]`<br>\nGiven a submission object, it downloads its file and returns it as a `bytes` object.\n\n* `userpage(user: str) -> Tuple[str, str, bs4.BeautifulSoup]`<br>\nReturns the user\'s full display name - i.e. with capital letters and extra characters such as "_" -, the user\'s status - the first character found beside the user name - and the parsed profile text in HTML.\n\n* `gallery(user: str, page: int = 1) -> Tuple[List[SubPartial], int]`<br>\nReturns the list of submissions found on a specific gallery page and the number of the next page. The returned page number is set to 0 if it is the last page.\n\n* `scraps(user: str, page: int = 1) -> -> Tuple[List[SubPartial], int]`<br>\nSame as `gallery()`, but scrapes a user\'s scraps page instead.\n\n* `favorites(user: str, page: str = \'\') -> Tuple[List[SubPartial], str]`<br>\nAs `gallery()` and `scraps()` it downloads a user\'s favorites page. Because of how favorites pages work on FA, the `page` argument (and the one returned) are strings. If the favorites page is the last then an empty string is returned as next page. An empty page value as argument is equivalent to page 1.<br>\n*Note:* favorites page "numbers" do not follow any scheme and are only generated server-side.\n\n* `search(q: str = \'\', page: int = 0, **params) -> Tuple[List[SubPartial], int, int, int, int]`<br>\nParses FA search given the query (and optional other params) and returns the submissions found and the next page together with basic search statistics: the number of the first submission in the page, the number of the last submission in the page (0-indexed), and the total number of submissions found in the search. For example if the the last three returned integers are 1, 47 and 437, then the the page contains submissions 1 through 48 of a search that has found a total of 437 submissions.<br>\n*Note:* as of 2020-08-01 the "/search" path is disallowed by FA\'s robots.txt.\n\n## SubPartial\n\nThis lightweight submission object is used to contain the information gathered when parsing gallery, scraps, favorites and search pages. It contains only the following fields:\n\n* `id` submission id\n* `title` submission title\n* `author` submission author\n* `rating` submission rating [general, mature, adult]\n* `type` submission type [text, image, etc...]\n\n`SubPartial` can be directly casted to a dict object or iterated through.\n\n### Init\n\n`__init__(figure_tag: bs4.element.Tag)`\n\n`SubPartial` init needs a figure tag taken from a parsed page. This tag is not saved in the submission object.\n\n### Methods\n\n* `parse_figure_tag(figure_tag: bs4.element.Tag)`<br>\nTakes a figure tag from a parsed page and parses it for information.\n\n## Sub\n\nThe main class that parses and holds submission metadata.\n\n* `id` submission id\n* `title` submission title\n* `author` submission author\n* `date` upload date in YYYY-MM-DD format\n* `tags` tags list\n* `category` category \\*\n* `species` species \\*\n* `gender` gender \\*\n* `rating` rating \\*\n* `description` the description as an HTML formatted string\n* `file_url` the url to the submission file\n\n\\* these are extracted exactly as they appear on the submission page\n\n`Sub` can be directly casted to a dict object and iterated through.\n\n### Init\n\n`__init__(self, sub_page: bs4.BeautifulSoup = None)`\n\nTo initialise the object, An optional `bs4.BeautifulSoup` object is needed containing the parsed HTML of a submission page.\n\nThe `sub_page` argument is saved as an instance variable and is then parsed to obtain the other fields.\n\nIf no `sub_page` is passed then the object fields will remain at their default - empty - value.\n\n### Methods\n\n* `parse_page()`<br>\nParses the stored submission page for metadata.\n\n## Examples\n\nTo see examples of the various functionalities of the module, use the scripts located in the examples folder.\nThese can be run directly, provided the requirements are met.\n\nSee the specific scripts for more details.\n',
    'author': 'Matteo Campinoti',
    'author_email': 'matteo.campinoti94@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/MatteoCampinoti94/FAAPI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
